searchData={"items":[{"type":"module","title":"KeyValidator","doc":"Compile-time validation to assure all the map/keyword keys exist in the target struct.\nUse case: maps that will be merged with structs.\n\nLibrary proivdes compile-time check macro for key validity of map/keyword keys for merge with structs.\n\nExposes the `KeyValidator.for_struct/2` macro.","ref":"KeyValidator.html"},{"type":"module","title":"Use cases - KeyValidator","doc":"The macro targets the situations where working with map/keyword literals that will be later cast onto the known structs.\n\nElixir and Ecto has built-in functions that perform the key validity check, but only at runtime:\n\n- `Kernel.struct!/2`\n- `Ecto.Query.API.merge/2`\n\nIn certain situations, the conformity between map/keyword keys can be checked already at the compile-time. One example is when we have present map/keyword **literals** in our code that we know ahead that will be used for casting onto structs. Let's take a look at the following example:\n\n```elixir\ndefmodule User do\ndefstruct name: \"john\"\nend\n\n# Following line is a runtime only check:\n\nKernel.struct!(User, %{name: \"Jakub\"})\n#=> %User{name: \"Jakub\"}\n\n# Runtime error on key typo:\n\nKernel.struct!(User, %{nam__e: \"Jakub\"})\n#=> ** (KeyError) key :nam__e not found\n```\n\nThe expression `Kernel.struct!(User, %{name: \"Jakub\"})` uses a map literal (`%{name: \"Jakub\"}`). Since the User struct module together with the map literal is defined at the compile time, we can leverage the power of compile-time macros to validate those. This is where `KeyValidator.for_struct/2` comes to help:\n\n```elixir\ndefmodule User do\ndefstruct name: \"john\"\nend\n\nimport KeyValidator\n\n# Succesfull validation. Returns the map:\n\nuser_map = for_struct(User, %{name: \"Jakub\"})\n#=> %{name: \"Jakub\"}\n\nKernel.struct!(User, user_map)\n#=> %User{name: \"Jakub\"}\n\n# Compile time error on \"nam__e:\" key typo\n\nuser_map2 = for_struct(User, %{nam__e: \"Jakub\"})\n#=>** (KeyError) Key :name_e not found in User\n```\n\nAs we can see `for_struct/2` macro allows some category of errors to be caught at very early stage in the development workflow. No need to wait the code to crash at runtime if there's a opportunity to check the key conformity before that. This is not a silver bullet though: the macro cannot accept dynamic variables, because their content cannot be evaluated during compilation.","ref":"KeyValidator.html#module-use-cases"},{"type":"module","title":"Extended example - KeyValidator","doc":"Useful to work with Ecto.Query.select_merge/3 when working with `virtual_fields`\n\n```elixir\ndefmodule Post do\nuse Ecto.Schema\nschema \"posts\" do\n field :author_firstname, :string\n field :author_lastname, :string\n field :author, :string, virtual_field: true\nend\nend\n\ndefmodule Posts do\nimport KeyValidator\n\ndef list_posts do\n Post\n |> select_merge([p], for_struct(Post, %{author: p.author_firstname <> \" \" <> p.author_lastname}))\n |> Repo.all()\nend\nend\n```\n\nThe following code will raise a Key Error with message: \"Key :author_non_existent_key not found in Post\"\n\n```elixir\ndefmodule Posts do\nimport KeyValidator\n\ndef list_posts_error do\nPost\n|> select_merge([p], for_struct(Post, %{author_non_existent_key: \"some value\"}))\n|> Repo.all()\n end\nend\n```","ref":"KeyValidator.html#module-extended-example"},{"type":"macro","title":"KeyValidator.for_struct/2","doc":"Validates all the map/keyword keys exist in the target struct at compile-time.\n\n  Raises compile-time errors if key does not exist.\n\n  - `module_or_struct` : Module atom (which defines defstruct) or struct (ex. %ModuleStruct{}).\n  - `fields` : map or keyword **literal**.\n\n\n  Returns `fields` when all the keys in`fields` are included in the target struct.\n\n  Raises:\n\n  - `KeyError` when any key in the fields is not found in struct.\n  - `ArgumentError` when\n    + `module` is not a module that defines struct\n    + `fields` are not a map/keyword literal","ref":"KeyValidator.html#for_struct/2"},{"type":"macro","title":"Examples - KeyValidator.for_struct/2","doc":"iex> import KeyValidator\n    iex> defmodule Post do\n          defstruct [:author]\n        end\n\n    iex> for_struct(Post, %{author: \"Jakub\"})\n    %{author: \"Jakub\"}\n\n    iex> for_struct(%Post{}, %{author: \"Jakub\"})\n    %{author: \"Jakub\"}\n\n    iex> for_struct(Post, author: \"Jakub\")\n    [author: \"Jakub\"]\n\n    iex> for_struct(Post, %{auth_typo_or: \"Jakub\"})\n    ** (KeyError) Key :auth_typo_or not found in Elixir.Post\n\n    iex> for_struct(ModuleWithNoStruct, %{author: \"Jakub\"})\n    ** (ArgumentError) Argument is not a module that defines a struct.","ref":"KeyValidator.html#for_struct/2-examples"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,48]}}